import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import copy
import math
import scipy.linalg

from algs.base_learner import kron

from utils import median_trick

from sklearn.kernel_approximation import RBFSampler

import multiprocessing



def get_inverse_sigma(replay_buffer, rbf_feature, lamb, feature_dim, h, queue):

    obses, actions, rewards, next_obses = replay_buffer.get_full(recent_size=100000)
    obsac = np.concatenate((obses.cpu(),actions.cpu()),axis=-1)
    #bandwidth = median_trick(obsac)
    phi = rbf_feature.fit_transform(obsac / 5)
    feature = torch.as_tensor(phi)
    Sigma = torch.matmul(feature.T, feature) + lamb * torch.eye(feature_dim)
    Sigma_invs = torch.inverse(Sigma)

    queue.put([h, Sigma_invs.numpy(), phi])

class LSVI_UCB(object): 

    def __init__(
        self,
        obs_dim,
        state_dim,
        action_dim,
        horizon,
        alpha,
        device,
        #rep_learners,
        variable_latent,
        env_temperature,
        feature_size = 1000,
        seed=12,
        lamb = 1
    ):

        self.obs_dim = obs_dim
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.horizon = horizon

        self.feature_dim = feature_size

        self.device = "cpu"
        #self.rep_learners = rep_learners

        self.lamb = lamb
        self.alpha = alpha

        self.W = torch.rand((self.horizon, self.feature_dim)).to(self.device)
        self.Sigma_invs = torch.zeros((self.horizon, self.feature_dim, self.feature_dim)).to(self.device)

        self.Q_max = torch.tensor(self.horizon)

        rotation = scipy.linalg.hadamard(obs_dim)
        self.A = torch.inverse(torch.as_tensor(rotation, device=self.device).float())

        self.variable_latent = variable_latent
        self.env_temperature = env_temperature

        self.rbf_feature = RBFSampler(gamma=1, random_state=seed, n_components=feature_size)

        self.bandwidth = np.ones(horizon) * 5

        print(self.alpha)
        print(self.env_temperature)
    
    def Q_values(self, obs, h):
        Qs = torch.zeros((len(obs),self.action_dim)).to(self.device)
        for a in range(self.action_dim):
            actions = torch.zeros((len(obs),self.action_dim)).to(self.device)
            actions[:,a] = 1
            feature = self.get_feature(torch.cat((obs, actions),dim=-1).cpu(),h)
            Q_est = torch.matmul(feature, self.W[h].to(self.device)) 
            ucb = torch.sqrt(torch.sum(torch.matmul(feature, self.Sigma_invs[h].to(self.device))*feature, 1))
            
            Qs[:,a] = torch.minimum(Q_est + self.alpha * ucb, self.Q_max)

        return Qs

    def act(self, obs, h):
        with torch.no_grad():
            obs = torch.FloatTensor(obs).to(self.device)
            obs = obs.unsqueeze(0)
            Qs = self.Q_values(obs, h)
            action = torch.argmax(Qs, dim=1)

        return action.cpu().data.numpy().flatten()

    def act_batch(self, obs, h):
        with torch.no_grad():
            obs = torch.FloatTensor(obs).to(self.device)
            Qs = self.Q_values(obs, h)
            action = torch.argmax(Qs, dim=1)

        return action.cpu().data.numpy().flatten()

    def get_feature(self, obsac, h):

        phi = self.rbf_feature.fit_transform(obsac / self.bandwidth[h])
        phi = torch.as_tensor(phi, device=self.device).float()

        return phi


    def update(self, buffers):
        assert len(buffers) == self.horizon
        queue = multiprocessing.Queue()
        workers = []
        
        cur_feature = {}
        
        for h in range(self.horizon):
            worker_args = (buffers[h], self.rbf_feature, self.lamb, self.feature_dim, h, queue)
            workers.append(multiprocessing.Process(target=get_inverse_sigma, args=worker_args))
        for worker in workers:
            worker.start()

        for _ in workers:
            pid, sigma_inverse, feature = queue.get()
            self.Sigma_invs[pid] = torch.as_tensor(sigma_inverse)
            cur_feature[pid] = torch.as_tensor(feature).float()

        for h in range(self.horizon)[::-1]:
            obses, actions, rewards, next_obses = buffers[h].get_full(recent_size=100000)
            if h == self.horizon - 1:
                target_Q = rewards
            else:
                Q_prime = torch.max(self.Q_values(next_obses, h+1),dim=1)[0].unsqueeze(-1)
                #print(Q_prime)
                #print(rewards)
                target_Q = rewards + Q_prime
            # print(target_Q.shape)
            # print(feature.shape)
            # print((feature * target_Q).shape)
            # print(torch.sum((feature * target_Q), 0).shape)
            # print(target_Q.shape)
            # print(feature.shape)
            self.W[h] = torch.matmul(self.Sigma_invs[h].to(self.device), torch.sum(cur_feature[h] * target_Q, 0))


        #print(self.bandwidth)

    def save_weight(self, path):
        for h in range(self.horizon):
            torch.save(self.W[h],"{}/W_{}.pth".format(path,str(h)))
            torch.save(self.Sigma_invs[h], "{}/Sigma_{}.pth".format(path,str(h)))

    def load_weight(self, path):
        for h in range(self.horizon):
            self.W[h] = torch.load("{}/W_{}.pth".format(path,str(h)))
            print(h)
            print(self.W[h].reshape((10,3)))
            self.Sigma_invs[h] = torch.load("{}/Sigma_{}.pth".format(path,str(h)))








